# -*- coding: utf-8 -*-
"""Models should all be enabled.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 15:51

"""
from typing import List, Tuple

from olivertwist.manifest import Manifest, Node
from olivertwist.rules.utils import partition


def no_disabled_models(manifest: Manifest) -> Tuple[List[Node], List[Node]]:
    passes, failures = partition(lambda x: not x.is_enabled, manifest.nodes())
    return list(passes), list(failures)



# -*- coding: utf-8 -*-
"""Models should have resolvable dependencies.

Copyright (C) 2020, Auto Trader UK
Created 16. Dec 2020 12:53

"""
from typing import List, Tuple

from olivertwist.manifest import Manifest, Node
from olivertwist.rules.utils import partition


def no_orphaned_models(manifest: Manifest) -> Tuple[List[Node], List[Node]]:
    """
    return [
        node for node in dbt_manifest_file['nodes'].values()
        if is_staging(node) or is_mart(node)
        if not node['depends_on']['nodes']
    ]
    """

    def is_orphan(node: Node) -> bool:
        dependencies = list(manifest.graph.predecessors(node.id))
        return (node.is_staging or node.is_mart) and len(dependencies) < 1

    passes, failures = partition(is_orphan, manifest.nodes())
    return list(passes), list(failures)



# -*- coding: utf-8 -*-
"""Document engine here.

Copyright (C) 2020, Auto Trader UK
Created 16. Dec 2020 14:45

"""
from typing import List

import networkx as nx

from olivertwist.manifest import Manifest
from olivertwist.metricengine.result import MetricResult


class MetricEngine:
    def run(self, manifest: Manifest) -> List[MetricResult]:
        graph = manifest.graph
        degree_centrality = nx.centrality.degree_centrality(graph)
        in_degree_centrality = nx.centrality.in_degree_centrality(graph)
        out_degree_centrality = nx.centrality.out_degree_centrality(graph)
        closeness_centrality = nx.centrality.closeness_centrality(graph)
        betweenness_centrality = nx.centrality.betweenness_centrality(graph)
        pagerank = nx.link_analysis.pagerank_alg.pagerank(graph)

        results = []
        for node in graph.nodes:
            results.append(
                MetricResult(
                    manifest.get_node(node),
                    degree_centrality[node],
                    in_degree_centrality[node],
                    out_degree_centrality[node],
                    closeness_centrality[node],
                    betweenness_centrality[node],
                    pagerank[node],
                )
            )

        return results



# -*- coding: utf-8 -*-
"""Document node here.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 14:56

"""
from copy import deepcopy
from typing import Dict

import networkx as nx


class Manifest:
    def __init__(self, data: Dict[str, Dict]):
        self._data = data
        self._graph = self._generate_graph(data)

    @property
    def graph(self) -> nx.DiGraph:
        return self._graph.copy()

    def nodes(self):
        for node_id in self.graph.nodes:
            yield self.get_node(node_id)

    def get_node(self, node_id: str) -> "Node":
        node_dicts = ["nodes", "exposures", "sources"]

        for node_dict in node_dicts:
            try:
                return Node(self._data[node_dict][node_id])
            except KeyError:
                continue

        for node in self._data["disabled"]:
            if node["unique_id"] == node_id:
                return Node(node)

        raise KeyError(f"node_id '{node_id}' not found")

    @staticmethod
    def _generate_graph(data) -> nx.DiGraph:
        graph = nx.DiGraph()

        for node_id, child_node_ids in data["child_map"].items():
            graph.add_node(node_id)
            for child_node_id in child_node_ids:
                graph.add_edge(node_id, child_node_id)

        return graph


class Node:
    def __init__(self, dbt_node_data: dict):
        self.data = dbt_node_data

    @property
    def id(self) -> str:
        return self.data["unique_id"]

    @property
    def is_enabled(self) -> bool:
        # FIXME: Feels hacky to assume no config means the model is enabled. This was added to make exposures work
        return self.data.get("config", {}).get("enabled", True)

    @property
    def is_mart(self) -> bool:
        return self.data["resource_type"] == "model" and self.__fqn_contains("marts")

    @property
    def is_source(self) -> bool:
        return self.data["resource_type"] == "source"

    @property
    def is_staging(self) -> bool:
        return self.data["resource_type"] == "model" and (
            self.__fqn_contains("staging") or "stg_" in self.id
        )

    def __fqn_contains(self, namespace: str) -> bool:
        return namespace in self.data.get("fqn", [])



import json
import os
import pathlib
import shutil

from jinja2 import Environment, PackageLoader

from olivertwist.reporter.model import *

root = pathlib.Path(__file__).parent
target_dir = pathlib.Path(os.getcwd()) / "target"

templates_dir = root / "templates"
css_dir = root / "html" / "css"
webfonts_dir = root / "html" / "webfonts"
images_dir = root / "html" / "images"
env = Environment(loader=PackageLoader("olivertwist.reporter", "templates"))
index_template = env.get_template("index.html")
model_template = env.get_template("model.html")


def output_json(oliver_twist_json):
    shutil.rmtree(target_dir, ignore_errors=True)
    os.mkdir(target_dir)
    filename = target_dir / "oliver_twist.json"
    with open(filename, "w") as outfile:
        json.dump(oliver_twist_json, outfile)


def render_html_report(oliver_twist_json):
    shutil.copytree(css_dir, target_dir / "css")
    shutil.copytree(webfonts_dir, target_dir / "webfonts")
    shutil.copytree(images_dir, target_dir / "images")

    filename = target_dir / "index.html"
    with open(filename, "w") as fh:
        fh.write(index_template.render(oliver_twist_json))

    for model in oliver_twist_json["models"]:
        filename = os.path.join(target_dir, model["model_name"] + ".html")
        with open(filename, "w") as fh:
            fh.write(model_template.render(model))



from typing import Dict

from olivertwist.manifest import Node
from olivertwist.metricengine.result import MetricResult
from olivertwist.reporter.model import *
from olivertwist.ruleengine.result import Result
from olivertwist.ruleengine.rule import Rule


def to_html_report(
    domain_results: List[Result], metric_results: List[MetricResult]
) -> Report:
    models: List[ReportModel] = __organise_by_model(domain_results, metric_results)
    summary: ReportSummary = __derive_summary(models)
    return Report(summary, models)


def __derive_summary(models: List[ReportModel]) -> ReportSummary:
    # TODO: summary.skipped?
    skipped = 0
    errored = len([a_model for a_model in models if a_model.summary.errored > 0])
    passed = len(models) - errored

    return ReportSummary(passed, skipped, errored)


def __organise_by_model(
    domain_results: List[Result], metric_results: List[MetricResult]
) -> List[ReportModel]:
    # model_key -> model
    deduped_nodes: Dict[str, Node] = __domain_node_by_model_name(domain_results)
    passed_rules_by_model: Dict[str, List[Rule]] = __passed_rules_by_model_name(
        domain_results
    )
    failed_rules_by_model: Dict[str, List[Rule]] = __failed_rules_by_model_name(
        domain_results
    )

    html_rules_by_model: Dict[str, List[ReportRule]] = __html_rules_by_model_name(
        passed_rules_by_model, failed_rules_by_model
    )
    metrics_results_by_model = __metric_results_by_model_name(metric_results)

    html_models = []
    for model_name in deduped_nodes.keys():
        html_summary = ReportSummary(
            len(passed_rules_by_model.get(model_name, [])),
            0,
            len(failed_rules_by_model.get(model_name, [])),
        )
        html_model = ReportModel(
            model_name,
            "",
            model_name,
            metrics_results_by_model[model_name],
            html_rules_by_model[model_name],
            html_summary,
        )
        html_models.append(html_model)

    return html_models


def __html_rules_by_model_name(
    passed_rules_by_model: Dict[str, List[Rule]],
    failed_rules_by_model: Dict[str, List[Rule]],
) -> Dict[str, List[ReportRule]]:
    result: Dict[str, List[ReportRule]] = {}
    for model_name, domain_rules in passed_rules_by_model.items():
        html_rules: List[ReportRule] = [
            ReportRule(domain_rule.name, domain_rule.message, ReportStatus.PASSED)
            for domain_rule in domain_rules
        ]
        result[model_name] = html_rules

    for model_name, domain_rules in failed_rules_by_model.items():
        html_rules: List[ReportRule] = [
            ReportRule(domain_rule.name, domain_rule.message, ReportStatus.ERRORED)
            for domain_rule in domain_rules
        ]
        result.setdefault(model_name, []).extend(html_rules)
    return result


def __passed_rules_by_model_name(domain: List[Result]) -> Dict[str, List[Rule]]:
    passed_rules: Dict[str, List[Rule]] = {}
    for a_domain in domain:
        for node in a_domain.passes:
            passed_rules.setdefault(node.id, []).append(a_domain.rule)
    return passed_rules


def __failed_rules_by_model_name(domain: List[Result]) -> Dict[str, List[Rule]]:
    failed_rules: Dict[str, List[Rule]] = {}
    for a_domain in domain:
        for node in a_domain.failures:
            failed_rules.setdefault(node.id, []).append(a_domain.rule)
    return failed_rules


def __domain_node_by_model_name(domain: List[Result]) -> Dict[str, Node]:
    deduped_nodes: Dict[str, Node] = {}
    for a_domain in domain:
        all_nodes: List[Node] = a_domain.passes + a_domain.failures
        for duped_node in all_nodes:
            deduped_nodes[duped_node.id] = duped_node
    return deduped_nodes


def __metric_results_by_model_name(
    metric_results: List[MetricResult],
) -> Dict[str, List[ReportMetrics]]:
    results = {}
    for result in metric_results:
        metrics = []
        metrics.append(ReportMetrics("degree_centrality", result.degree_centrality))
        metrics.append(
            ReportMetrics("in_degree_centrality", result.in_degree_centrality)
        )
        metrics.append(
            ReportMetrics("out_degree_centrality", result.out_degree_centrality)
        )
        metrics.append(
            ReportMetrics("closeness_centrality", result.closeness_centrality)
        )
        metrics.append(
            ReportMetrics("betweenness_centrality", result.betweenness_centrality)
        )
        metrics.append(ReportMetrics("pagerank", result.pagerank))
        results[result.node.id] = metrics
    return results



# -*- coding: utf-8 -*-
"""Models should not be rejoined downstream

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 17:40

"""
from itertools import product
from typing import List, Tuple

from networkx import DiGraph, all_simple_paths

from olivertwist.manifest import Manifest, Node
from olivertwist.rules.utils import partition


def __find_rejoin_nodes(digraph: DiGraph):
    source_nodes = []
    target_nodes = []
    for n in list(digraph.nodes):
        if digraph.in_degree(n) == 0 and digraph.out_degree(n) >= 1:
            source_nodes.append(n)
        if digraph.in_degree(n) >= 1 and digraph.out_degree(n) == 0:
            target_nodes.append(n)

    rejoin_nodes = set()
    for s, t in list(product(source_nodes, target_nodes)):
        possible_paths = []
        for path in all_simple_paths(digraph, source=s, target=t):
            possible_paths.append(path)
        if len(possible_paths) > 1:
            # rejoin as there is more than one path between two nodes
            tmp = DiGraph()
            for paths in possible_paths:
                # build a graph from the possible paths
                for i in range(1, len(paths)):
                    tmp.add_edge(paths[i - 1], paths[i])
            for node in list(tmp.nodes):
                # rejoin nodes must have an out degree greater than one
                if tmp.out_degree(node) > 1:
                    rejoin_nodes.add(node)
    return rejoin_nodes


def no_rejoin_models(
    manifest: Manifest,
) -> Tuple[List[Node], List[Node]]:
    rejoin_nodes = __find_rejoin_nodes(manifest.graph)

    def is_rejoin_node(node: Node):
        return node.id in rejoin_nodes

    passes, failures = partition(is_rejoin_node, manifest.nodes())
    return list(passes), list(failures)



# -*- coding: utf-8 -*-
"""Document result here.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 17:09

"""
from typing import List

from olivertwist.manifest import Node
from olivertwist.ruleengine.rule import Rule


class Result:
    def __init__(self, rule: Rule, passes: List[Node], failures: List[Node]):
        self.rule = rule
        self.passes = passes
        self.failures = failures

    @property
    def has_failures(self):
        return bool(self.failures)



# -*- coding: utf-8 -*-
"""Staging models should have a single source.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 17:40

"""

from typing import List, Tuple

from olivertwist.manifest import Manifest, Node
from olivertwist.rules.utils import partition


def staging_models_have_single_source(
    manifest: Manifest,
) -> Tuple[List[Node], List[Node]]:
    def staging_model_has_more_than_one_source(node: Node):
        sources = [
            p
            for p in manifest.graph.predecessors(node.id)
            if manifest.get_node(p).is_source
        ]
        return node.is_staging and len(list(sources)) > 1

    passes, failures = partition(
        staging_model_has_more_than_one_source, manifest.nodes()
    )
    return list(passes), list(failures)



from dataclasses import dataclass
from enum import Enum
from json import JSONEncoder
from typing import List


class MyEncoder(JSONEncoder):
    def default(self, o):
        return o.__dict__


@dataclass
class ReportStatus(str, Enum):
    PASSED = "passed"
    SKIPPED = "skipped"
    ERRORED = "errored"


@dataclass
class ReportRule:
    def __init__(
        self,
        name: str,
        message: str,
        status: ReportStatus,
    ):
        self.name = name
        self.message = message
        self.status = status


@dataclass
class ReportMetrics:
    def __init__(
        self,
        name: str,
        score: int,
    ) -> None:
        self.name = name
        self.score = score
        self.pretty_name = name.replace("_", " ").capitalize()


@dataclass
class ReportSummary:
    def __init__(
        self,
        passed: int,
        skipped: int,
        errored: int,
    ) -> None:
        self.passed = passed
        self.skipped = skipped
        self.errored = errored


@dataclass
class ReportModel:
    def __init__(
        self,
        model_key: str,
        file_path: str,
        model_name: str,
        metrics: List[ReportMetrics],
        rules: List[ReportRule],
        summary: ReportSummary,
    ) -> None:
        self.model_key = model_key
        self.file_path = file_path
        self.model_name = model_name
        self.metrics = metrics
        self.rules = rules
        self.summary = summary


@dataclass
class Report:
    def __init__(self, summary: ReportSummary, models: List[ReportModel]) -> None:
        self.summary = summary
        self.models = models



# -*- coding: utf-8 -*-
"""Document rule here.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 14:31

"""
from typing import Callable, List, Tuple

from olivertwist.manifest import Manifest, Node


class Rule:
    def __init__(
        self,
        name: str,
        message: str,
        func: Callable[[Manifest], Tuple[List[Node], List[Node]]],
    ):
        self.name = name
        self.message = message
        self.func = func

    def apply(self, manifest) -> Tuple[List[Node], List[Node]]:
        return self.func(manifest)



# -*- coding: utf-8 -*-
"""Main entry point for olivertwist.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 15:32

"""
import json
import logging
import os
import webbrowser
from typing import List

import click

from olivertwist.manifest import Manifest
from olivertwist.metricengine.engine import MetricEngine
from olivertwist.reporter.adapter import to_html_report
from olivertwist.reporter.model import MyEncoder
from olivertwist.reporter.reporter import output_json, render_html_report
from olivertwist.ruleengine.engine import RuleEngine
from olivertwist.ruleengine.result import Result

logger = logging.getLogger("olivertwist")


@click.command()
@click.argument("input", type=click.File("r"))
@click.option("--html/--no-html", default=True, help="Do/Don't output report in HTML")
@click.option(
    "--browser/--no-browser",
    default=False,
    help="Do/Don't open HTML report in browser. Implies --html",
)
def main(input, html=True, browser=False):
    manifest = Manifest(json.load(input))
    rule_engine = RuleEngine.with_default_rules()
    results = rule_engine.run(manifest)
    format_for_terminal(results)
    metric_results = MetricEngine().run(manifest)
    report = to_html_report(results, metric_results)
    oliver_twist = json.loads(MyEncoder().encode(report))
    output_json(oliver_twist)
    if html or browser:
        logger.debug("Generating HTML report...")
        render_html_report(oliver_twist)
        if browser:
            webbrowser.open(f"file://{os.getcwd()}/target/index.html")

    exit_message(results)


def format_for_terminal(results: List[Result]):
    for result in results:
        colour = "red" if result.has_failures else "green"
        click.secho(f"{result.rule.name}:", fg=colour)
        for node in result.failures:
            click.secho(f" - {node.id}", fg="red")

        click.echo()


def exit_message(results: List[Result]):
    if any([result.has_failures for result in results]):
        click.get_current_context().exit("🔀 Twisted!")
    else:
        click.echo("🟢 Oliver (all of your) models look good!")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""Useful utilities for writing rules.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 17:50

"""
from itertools import filterfalse, tee


def partition(pred, iterable):
    "Use a predicate to partition entries into false entries and true entries."
    t1, t2 = tee(iterable)
    return filterfalse(pred, t1), filter(pred, t2)



# -*- coding: utf-8 -*-
"""Document rule here.

Copyright (C) 2020, Auto Trader UK
Created 16. Dec 2020 14:31

"""
from typing import Callable, List, Tuple
from olivertwist.manifest import Node


class MetricResult:
    def __init__(
        self,
        node: Node,
        degree_centrality: float,
        in_degree_centrality: float,
        out_degree_centrality: float,
        closeness_centrality: float,
        betweenness_centrality: float,
        pagerank: float,
    ):
        self.node = node
        self.degree_centrality = degree_centrality
        self.in_degree_centrality = in_degree_centrality
        self.out_degree_centrality = out_degree_centrality
        self.closeness_centrality = closeness_centrality
        self.betweenness_centrality = betweenness_centrality
        self.pagerank = pagerank



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['olivertwist',
 'olivertwist.metricengine',
 'olivertwist.reporter',
 'olivertwist.ruleengine',
 'olivertwist.rules']

package_data = \
{'': ['*'],
 'olivertwist.reporter': ['html/css/*',
                          'html/images/*',
                          'html/webfonts/*',
                          'templates/*']}

install_requires = \
['Jinja2==2.11.2',
 'click>=7.1.2,<8.0.0',
 'colorama>=0.4.4,<0.5.0',
 'networkx>=2.5,<3.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.8,<0.9']}

entry_points = \
{'console_scripts': ['olivertwist = olivertwist.main:main',
                     'ot = olivertwist.main:main']}

setup_kwargs = {
    'name': 'olivertwist',
    'version': '0.1.2',
    'description': 'DBT DAG Auditor',
    'long_description': '\n\n![Alt text](./images/oliver_twist_logo.png)\n# oliver-twist\n\nDAG Auditor\n\n![Build status badge](https://github.com/autotraderuk/oliver-twist/workflows/CI/badge.svg)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\noliver-twist is a dag auditing tool that audits the [DBT](https://www.getdbt.com/) DAG and generates a summary report. The rules implemented can be found [here](RULES.md)\n\n![please sir, can I automate my DAG auditing](./images/oliver_dag_meme.jpg)\n\n# Getting Started\n\nTo get started, install the package\n\n```shell\n$ pip install olivertwist\n```\n\nand then run it by passing it your dbt manifest JSON\n\n```shell\nolivertwist manifest.json\n```\n\nThis will report any failures to the console, and also in HTML format in a directory called `target`. You can optionally auto-open the report in a browser with:\n\n```shell\nolivertwist manifest.json --browser\n```\n\nFull options are available with:\n\n\n```shell\nolivertwist manifest.json --help\n```\n\n## Developer\n\n### To dev locally\n\nClone this repo and install all the projects packages:\n\n`poetry install`\n\nTo get the latest versions of the dependencies and to update the poetry.lock file run:\n\n`poetry update`\n\nTo run oliver-twist and generate the summary report run:\n\n`poetry run olivertwist example_manifest.json`\n\n\n### Creating a distribution\n\n```poetry build --format wheel```\n',
    'author': 'Angelos Georgiadis',
    'author_email': 'angelos.georgiadis@autotrader.co.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/autotraderuk/oliver-twist',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""Document engine here.

Copyright (C) 2020, Auto Trader UK
Created 15. Dec 2020 14:45

"""
from typing import List

from olivertwist.manifest import Manifest
from olivertwist.ruleengine.rule import Rule
from olivertwist.ruleengine.result import Result
from olivertwist.rules.no_disabled_models import no_disabled_models
from olivertwist.rules.no_orphaned_models import no_orphaned_models
from olivertwist.rules.staging_has_single_source import (
    staging_models_have_single_source,
)
from olivertwist.rules.no_rejoin_models import no_rejoin_models


class RuleEngine:
    def __init__(self, rules: List[Rule]):
        self.rules = rules

    @classmethod
    def with_default_rules(cls) -> "RuleEngine":
        return cls(
            [
                Rule(
                    name="Disabled models",
                    message="If you have disabled a script that you no longer require, you should delete it.",
                    func=no_disabled_models,
                ),
                Rule(
                    name="No orphaned models",
                    message="Staging or mart model(s) contain zero references.",
                    func=no_orphaned_models,
                ),
                Rule(
                    name="Staging scripts referencing multiple sources",
                    message="A staging script is referencing multiple sources.",
                    func=staging_models_have_single_source,
                ),
                Rule(
                    name="Rejoin models",
                    message="A model is being rejoined downstream",
                    func=no_rejoin_models,
                ),
            ]
        )

    def run(self, manifest: Manifest) -> List[Result]:
        return [Result(rule, *rule.apply(manifest)) for rule in self.rules]
